import base64
import logging
import os
import re
from pathlib import Path
from typing import AsyncIterator, Optional

from qbee_gpio.metadata.interface import NowPlayingPoller
from qbee_gpio.metadata.model import NowPlaying
from qbee_gpio.metadata.pipe_reader import PipeReader

logger = logging.getLogger(__name__)


class LibrespotNowPlayingPoller(PipeReader, NowPlayingPoller):
    """Poll librespot metadata pipe and expose minimal information."""

    RE_METADATA = re.compile(r"artist:(.*?),album:(.*?),title:(.*)")

    def __init__(self, path: Path):
        super().__init__(path, b"\t")

    async def __aenter__(self):
        self.path.unlink(missing_ok=True)
        os.mkfifo(str(self.path))
        self.callback(self.path.unlink, missing_ok=True)
        return await super().__aenter__()

    async def poll(self) -> AsyncIterator[NowPlaying]:
        async with self:
            logger.debug("started now playing poller")
            last_event: Optional[NowPlaying] = None
            async for data in self._receive():
                if (event := self._decode(data)) and event != last_event:
                    logger.debug("now playing %r", event)
                    last_event = event
                    yield event

    def _decode(self, data: bytes) -> Optional[NowPlaying]:
        if match := self.RE_METADATA.search(data.decode()):
            artist, album, title = tuple(
                base64.b64decode(e).decode() for e in match.groups()
            )
            return NowPlaying(
                artist=", ".join(artist.strip().split("\n")),
                album=album,
                title=title,
            )
        return None



"""
Minimal GPIO types for static typing and mocking in tests.
"""

from typing import Literal, Union

Bit = Union[bool, Literal[0, 1]]


class GPIO:
    OUT = 0
    BCM = 0

    @classmethod
    def setmode(cls, mode: int): ...

    @classmethod
    def setup(cls, pin: int, mode: int): ...

    @classmethod
    def output(cls, pin: int, value: Bit): ...

    @classmethod
    def cleanup(cls): ...



import contextlib
from abc import ABC, abstractmethod
from typing import AsyncIterator

from qbee_gpio.metadata.model import NowPlaying


class NowPlayingPoller(contextlib.AbstractAsyncContextManager, ABC):
    """Poll metadata and expose minimal information."""

    @abstractmethod
    async def poll(self) -> AsyncIterator[NowPlaying]:
        yield NowPlaying("", "", "")



import asyncio
import contextlib
import unicodedata
from dataclasses import dataclass, field
from typing import Callable, Literal, Sequence, Tuple

try:
    from RPi import GPIO
except ModuleNotFoundError:
    from qbee_gpio.gpio.types import GPIO
from qbee_gpio.gpio.types import Bit


@dataclass
class GPIOLCDDisplay(contextlib.AbstractAsyncContextManager):
    """Hitachi HD44780 LCD controller.
    High level function to display text on the LCD.

    For more information see:
    - Condensed version: https://en.wikipedia.org/wiki/Hitachi_HD44780_LCD_controller
    - Datasheet: https://cdn-shop.adafruit.com/datasheets/HD44780.pdf
    """

    # PIN setup.
    # You will have to set up board/BCM mode yourself and those numbers should reflect that choice.
    pin_register_select: int
    pin_enable: int
    pin_data_4: int
    pin_data_5: int
    pin_data_6: int
    pin_data_7: int
    # Number of characters per line.
    width: int = 16
    # Number of lines.
    lines: Literal[1, 2, 4] = 2
    line_height: Literal[8, 10] = 8

    _line_addresses: Tuple[int, ...] = field(init=False)
    _lock: asyncio.Lock = field(default_factory=asyncio.Lock)
    _init: bool = False  # Has the LCD been initialized.

    def __post_init__(self):
        # Set up the line addresses (based on width for more than 2 lines).
        self._line_addresses = (0x00, 0x40, 0x00 + self.width, 0x40 + self.width)
        # Init PINs.
        for pin in (
            self.pin_register_select,
            self.pin_enable,
            self.pin_data_4,
            self.pin_data_5,
            self.pin_data_6,
            self.pin_data_7,
        ):
            GPIO.setup(pin, GPIO.OUT)

    async def __aenter__(self):
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        async with self._lock:
            self._init = False

    async def init(self) -> None:
        async with self._lock:
            # Init LCD.
            GPIO.output(self.pin_register_select, False)
            # No need to wait here as if the PI is booted power is already high enough.
            # Send 3 times the same command to ensure 8-bit mode.
            await self._send_command((0, 0, 1, 1))
            await asyncio.sleep(0.01)  # Wait more than 4.1ms here as per specs.
            await self._send_command((0, 0, 1, 1))
            await asyncio.sleep(0.001)  # Wait more than 100µs here as per specs.
            await self._send_command((0, 0, 1, 1))
            # Now switch to 4-bit.
            await self._send_command((0, 0, 1, 0))
            # Function set.
            await self._send_command(
                (0, 0, 1, 0),
                (
                    bool(self.lines > 1),
                    bool(self.line_height != 8),
                    0,
                    0,
                ),
            )
            # Display on/off control.
            await self._send_command(
                (0, 0, 0, 0),
                (
                    1,
                    1,  # Screen on.
                    0,  # Cursor off.
                    0,  # Cursor blink off.
                ),
            )
            # Clear display.
            await self._send_command((0, 0, 0, 0), (0, 0, 0, 1))
            # Entry mode set.
            await self._send_command(
                (0, 0, 0, 0),
                (
                    0,
                    1,
                    1,  # Cursor moves right.
                    0,  # Display does not shit.
                ),
            )
            self._init = True

    async def display(
        self,
        message: str,
        *,
        # Wrap content to spread on all available lines.
        wrap: bool = False,
        # This should be a function that takes a string and the width of the display
        # as arguments and return a string whose length is the same as the display.
        align: Callable[[str, int], str] = str.ljust,
    ) -> None:
        """Display a message on the screen."""
        if not message.strip():
            await self._send_command((0, 0, 0, 0), (0, 0, 0, 1))
            return
        lines = message.split("\n")
        if wrap:
            wrapped_lines = []
            for line in lines:
                # Convert each line in multiple ones based on the display width.
                wrapped_lines += [
                    line[i : i + self.width] for i in range(0, len(line), self.width)
                ]
            lines = wrapped_lines
        # Only keep lines we can display.
        lines = lines[: self.lines]
        # Add empty lines if needed.
        if len(lines) != self.lines:
            lines += [""] * (self.lines - len(lines))
        # Remove accents and align each line.
        lines = [
            align(remove_accents(line[: self.width]), self.width) for line in lines
        ]
        await self._print_lines(lines)

    async def _print_lines(self, lines: Sequence[str]) -> None:
        async with self._lock:
            if not self._init:
                return
            for i, line in enumerate(lines):
                await self._send_command(
                    *get_bits(0x80 + self._line_addresses[i]),
                )
                for char in line:
                    await self._send_char(ord(char))

    async def _send_char(self, byte: int) -> None:
        """Send a single byte."""
        # Set the mode.
        GPIO.output(self.pin_register_select, True)
        # Send bits 4-7 then 0-3.
        for bits in get_bits(byte):
            await self._send_4_bits(*bits)

    async def _send_command(
        self,
        *high_and_low_bits: Tuple[Bit, Bit, Bit, Bit],
    ) -> None:
        GPIO.output(self.pin_register_select, False)
        for bits in high_and_low_bits:
            await self._send_4_bits(*bits)
        if high_and_low_bits == ((0, 0, 0, 0), (0, 0, 0, 1)):
            # Wait more than 1.52ms after clear command.
            await asyncio.sleep(0.002)
        else:
            # Wait more than 37µs after other commands.
            await asyncio.sleep(0.00005)

    async def _send_4_bits(
        self,
        bit3: Bit,
        bit2: Bit,
        bit1: Bit,
        bit0: Bit,
    ) -> None:
        GPIO.output(self.pin_data_4, bit0)
        GPIO.output(self.pin_data_5, bit1)
        GPIO.output(self.pin_data_6, bit2)
        GPIO.output(self.pin_data_7, bit3)
        await self._enable()

    async def _enable(self) -> None:
        GPIO.output(self.pin_enable, True)
        await asyncio.sleep(0.000001)
        GPIO.output(self.pin_enable, False)


def get_bits(
    byte: int,
) -> Tuple[Tuple[Bit, Bit, Bit, Bit], Tuple[Bit, Bit, Bit, Bit]]:
    high_bits = byte >> 4
    return (
        (
            bool(high_bits & 0x08),
            bool(high_bits & 0x04),
            bool(high_bits & 0x02),
            bool(high_bits & 0x01),
        ),
        (
            bool(byte & 0x08),
            bool(byte & 0x04),
            bool(byte & 0x02),
            bool(byte & 0x01),
        ),
    )


def remove_accents(text: str) -> str:
    """Remove accents from text."""
    return (
        unicodedata.normalize("NFKD", text)
        .encode("ASCII", "ignore")
        .decode("utf-8", "ignore")
    )


async def debug():
    GPIO.setmode(GPIO.BCM)
    try:
        lcd = GPIOLCDDisplay(
            pin_register_select=int(input("pin register select: ")),
            pin_enable=int(input("pin enable: ")),
            pin_data_4=int(input("pin data 4: ")),
            pin_data_5=int(input("pin data 5: ")),
            pin_data_6=int(input("pin data 6: ")),
            pin_data_7=int(input("pin data 7: ")),
        )
        async with lcd:
            while True:
                await lcd.display(input("message: "), align=str.center)
    finally:
        GPIO.cleanup()


if __name__ == "__main__":
    asyncio.run(debug())



from functools import partial
from typing import Callable

from qbee_gpio.gpio.types import Bit

try:
    from RPi import GPIO
except ModuleNotFoundError:
    from qbee_gpio.gpio.types import GPIO


def gpio_switch(pin: int) -> Callable[[Bit], None]:
    GPIO.setup(pin, GPIO.OUT)
    return partial(GPIO.output, pin)



import base64
import logging
import re
from pathlib import Path
from typing import AsyncIterator, Optional

from qbee_gpio.metadata.interface import NowPlayingPoller
from qbee_gpio.metadata.model import NowPlaying
from qbee_gpio.metadata.pipe_reader import PipeReader

logger = logging.getLogger(__name__)


class ShairportNowPlayingPoller(PipeReader, NowPlayingPoller):
    """Poll shairport metadata pipe and expose minimal information."""

    def __init__(self, path: Path):
        super().__init__(
            path,
            f"<item><type>{b'ssnc'.hex()}</type><code>{b'mden'.hex()}</code>".encode(),
        )

    async def poll(self) -> AsyncIterator[NowPlaying]:
        async with self:
            logger.debug("started now playing poller")
            last_event: Optional[NowPlaying] = None
            async for data in self._receive():
                if (event := self._decode(data)) and event != last_event:
                    logger.debug("now playing %r", event)
                    last_event = event
                    yield event

    @staticmethod
    def _get_metadata(data: bytes, code: bytes) -> Optional[str]:
        match = re.search(
            f"<item><type>{b'core'.hex()}</type><code>{code.hex()}</code><length>[0-9]+</length>\n"
            '<data encoding="base64">\n(.*?)</data></item>',
            data.decode(),
        )
        if not match:
            return None
        return base64.b64decode(match.groups()[0]).decode()

    def _decode(self, data: bytes) -> Optional[NowPlaying]:
        if (
            (artist := self._get_metadata(data, b"asar"))
            and (album := self._get_metadata(data, b"asal"))
            and (title := self._get_metadata(data, b"minm"))
        ):
            return NowPlaying(
                artist=artist,
                album=album,
                title=title,
            )
        return None



from dataclasses import dataclass


@dataclass
class NowPlaying:
    artist: str
    album: str
    title: str

    def display(self, lines: int) -> str:
        if lines == 1:
            return self.title
        elif lines == 2:
            return f"{self.artist}\n{self.title}"
        else:
            return f"{self.artist}\n{self.album}\n{self.title}"



import asyncio
import contextlib
import logging
import os
from pathlib import Path
from typing import AsyncIterator, Optional

logger = logging.getLogger(__name__)


class PipeReader(contextlib.AsyncExitStack, asyncio.Protocol):
    """Read a pipe and expose an async iterator over data."""

    def __init__(self, path: Path, chunk_separator: bytes):
        super().__init__()
        self.path = path
        self._chunk_separator = chunk_separator
        # We need a queue as `data_received` is not async.
        self._received_data: asyncio.Queue[bytes] = asyncio.Queue()
        self._transport: Optional[asyncio.ReadTransport] = None

    async def __aenter__(self):
        await self._create_transport()
        self.callback(self._close_transport)
        return self

    async def _create_transport(self) -> None:
        pipe = self.enter_context(
            open(
                os.open(str(self.path), os.O_RDONLY | os.O_NONBLOCK),
            ),
        )
        self._transport, _ = await asyncio.get_running_loop().connect_read_pipe(
            lambda: self,
            pipe,
        )

    def _close_transport(self) -> None:
        if self._transport:
            self._transport.close()

    def data_received(self, data: bytes) -> None:
        self._received_data.put_nowait(data)

    def connection_lost(self, exc):
        # If no exception it means write end has closed.
        if exc:
            logger.warning("error received while reading %s: %r", self.path, exc)
        # Signal to reconnect.
        self._received_data.put_nowait(b"")

    async def _receive(self) -> AsyncIterator[bytes]:
        buffer = bytearray()
        while True:
            data = await self._received_data.get()
            if not data:
                # Reconnect.
                logger.debug("reconnecting reader of %s", self.path)
                await self._create_transport()
                continue
            buffer += data
            try:
                sep_index = buffer.index(self._chunk_separator)
            except ValueError:
                continue
            yield buffer[:sep_index]
            buffer = buffer[sep_index + len(self._chunk_separator) :]



import asyncio
import contextlib
import logging.config
from typing import List

from concurrent_tasks import BackgroundTask, LoopExceptionHandler

from qbee_gpio.config import QbeeConfig
from qbee_gpio.gpio import GPIOLCDDisplay, gpio_switch
from qbee_gpio.metadata import (
    LibrespotNowPlayingPoller,
    NowPlayingPoller,
    ShairportNowPlayingPoller,
)
from qbee_gpio.sound_poller import SoundPoller

logger = logging.getLogger(__name__)


class QbeeOrchestrator(contextlib.AsyncExitStack):
    """Orchestrate power, sound driver, metadata and display.

    This will run two main background tasks:
    - sound driver activity
    - now playing metadata polling and display update

    When activity is detected, everything is turned on and what is playing is displayed.
    When activity stops, a standby timer starts to turn off if no activity is detected in the meantime.
    """

    def __init__(self, debug: bool = False):
        super().__init__()
        self._cfg = QbeeConfig.load()
        logging.config.dictConfig(self._cfg.log_config)
        if debug:
            logging.getLogger().setLevel(logging.DEBUG)
        self._lcd = (
            GPIOLCDDisplay(
                pin_register_select=self._cfg.lcd.pin_register_select,
                pin_enable=self._cfg.lcd.pin_enable,
                pin_data_4=self._cfg.lcd.pin_data_4,
                pin_data_5=self._cfg.lcd.pin_data_5,
                pin_data_6=self._cfg.lcd.pin_data_6,
                pin_data_7=self._cfg.lcd.pin_data_7,
                width=self._cfg.lcd.width,
                lines=self._cfg.lcd.lines,
            )
            if self._cfg.lcd.enable
            else None
        )
        self._on_switch = (
            gpio_switch(self._cfg.sound_detection.pin_on)
            if self._cfg.sound_detection.enable
            else None
        )
        self._standby_switch = (
            gpio_switch(self._cfg.sound_detection.pin_standby)
            if self._cfg.sound_detection.enable
            else None
        )
        self._standby_task = (
            BackgroundTask(self._standby) if self._cfg.sound_detection.enable else None
        )
        self._sound_poller = (
            SoundPoller(
                self._cfg.sound_detection.driver_path,
            )
            if self._cfg.sound_detection.enable
            else None
        )

        self._poll_sound_task = (
            BackgroundTask(self._poll_sound) if self._sound_poller else None
        )
        self._poll_now_playing_tasks: List[BackgroundTask] = []
        if self._cfg.lcd.enable and self._cfg.lcd.shairport_metadata_path:
            self._poll_now_playing_tasks.append(
                BackgroundTask(
                    self._poll_now_playing,
                    ShairportNowPlayingPoller(self._cfg.lcd.shairport_metadata_path),
                )
            )
        if self._cfg.lcd.enable and self._cfg.lcd.librespot_metadata_path:
            self._poll_now_playing_tasks.append(
                BackgroundTask(
                    self._poll_now_playing,
                    LibrespotNowPlayingPoller(self._cfg.lcd.librespot_metadata_path),
                )
            )
        self._last_message = self._cfg.lcd.startup_message
        self._stop_event = asyncio.Event()

    async def __aenter__(self):
        self._switch(False)
        self.callback(self._switch, False)
        if self._standby_task:
            self.enter_context(self._standby_task)
        if self._poll_sound_task:
            self.enter_context(self._poll_sound_task)
        if self._lcd:
            await self.enter_async_context(self._lcd)
        for task in self._poll_now_playing_tasks:
            self.enter_context(task)
        return self

    async def run(self) -> None:
        logger.debug("starting orchestrator...")
        async with LoopExceptionHandler(stop_func=self._stop):
            async with self:
                logger.info("started orchestrator")
                await self._stop_event.wait()
        logger.debug("stopped orchestrator")

    async def _stop(self) -> None:
        logger.debug("stopping orchestrator...")
        self._stop_event.set()

    # Sound detection.

    async def _poll_sound(self) -> None:
        assert self._sound_poller
        assert self._standby_task
        async for output in self._sound_poller.poll():
            if output:
                logger.debug("cancelling standby mode")
                self._standby_task.cancel()
                self._switch(True)
                await self._print(self._last_message)
            else:
                await self._print("")  # Clear the display.
                self._standby_task.create()

    async def _standby(self) -> None:
        assert self._cfg.sound_detection
        if self._cfg.sound_detection.standby_duration is not None:
            logger.debug("entering standby mode")
            await asyncio.sleep(self._cfg.sound_detection.standby_duration)
            logger.debug("exiting standby mode")
            self._switch(False)
        if self._cfg.sound_detection.shutdown_command:
            logger.info("shutting down system...")
            await asyncio.create_subprocess_shell(
                self._cfg.sound_detection.shutdown_command,
            )

    # LCD.

    async def _poll_now_playing(self, poller: NowPlayingPoller) -> None:
        async for event in poller.poll():
            self._last_message = event.display(self._cfg.lcd.lines)
            await self._print(self._last_message)

    async def _print(self, message: str) -> None:
        if not self._lcd:
            return
        await self._lcd.display(message, align=str.center)

    # Switches.

    def _switch(self, value: bool) -> None:
        if not self._standby_switch or not self._on_switch:
            return
        logger.debug("turning %s", "on" if value else "off")
        self._on_switch(value)
        self._standby_switch(not value)
